# This module provides basic dialogs of Alicq
#
# Author: Igor Vergeichik

package require Tk
package require BWidget

namespace eval meta {
	set author "Ihar Viarheichyk <iverg@mail.ru>"
	set description "Basic implementation of chat dialogs."
	array set popup {
		type boolean default 0 property Global:UI:Messages|Popup
		description "Popup dialog window on new messages" save exit
	}
	array set tabbing {
		type boolean default 0 save exit
		description "Use tabbed panels within single window"
	}
	array set style {
		type variant save change description
		"Formatting style\n(can be overrided with per-window resources)"
	}
}

# Main menu methods (temporary) '''1
method Me::menu {Interface Tabbing} {args} {
	variable tabbing
	eval [list set tabbing] $args
}

method Me::menu {Interface {Popup Dialogs}} {args} {
	variable popup
	eval [list set popup] $args
}
#```1

#option add *messageStyle Chat interactive
option add *timeFormat %X widgetDefault
option add *dateFormat %x widgetDefault
option add *Text.underlineTag alias widgetDefault
option add *Text.headerBackground gray85 widgetDefault
option add *Text.bodyColor black widgetDefault

option add *EditBar.send.text [mc Send] widgetDefault
option add *EditBar.send.default active widgetDefault
option add *EditBar.close.relief flat widgetDefault
option add *EditBar.type.direction above widgetDefault
option add *EditBar.type.anchor w widgetDefault
option add *EditBar.AuxBar.cite.text [mc Cite] widgetDefault
option add *Tablist.Radiobutton.selectColor gray95 widgetDefault
option add *Tablist.Radiobutton.foreground black widgetDefault
option add *Tablist.Radiobutton.highlight blue widgetDefault

foreach {item value} {incoming	darkMagenta	outgoing	red
	client	darkBlue	server	darkBlue	sent	darkGray
	authrequest brown action blue} {
	option add *${item}Color $value widgetDefault
}


event add <<Send>> <Control-Return>
event add <<Cite>> <Control-q>
event add <<History>> <Control-h>
event add <<NextTab>> <Control-Right>
event add <<PrevTab>> <Control-Left>
foreach x {1 2 3 4 5 6 7 8 9} { event add <<ToTab>> <Control-Key-$x> }

# Create toplevel with saveable geometry
proc Toplevel {name class} {
	toplevel $name -class $class
	bindtags $name [concat [bindtags $name] SaveGeometry]
}
namespace export Toplevel
namespace eval :: [list namespace import [namespace current]::Toplevel]

# Raise windows, but not take focus if another dialog window has it
proc Raise {win {flag 0}} {
	set focus [focus]
	if {$focus!=""} { set focus [winfo toplevel $focus] }
	if {$focus=="" || $focus=="." || $flag} {
		raise $win
		focus $win
	} elseif {$focus!=$win} { lower $win $focus }
}


# Chat dialog for text messages used in single window mode '''1
namespace eval SingleChat {
	proc initialize {flag args} {
		upvar 1 self self
		set top [top]
		set focus [focus]
		[namespace parent]::Toplevel $top AlicqUserWindow
		if {$flag} { focus $top } elseif {$focus != ""} {
			set focus [winfo toplevel $focus]
			if {[winfo class $focus] == [winfo class $top]} {
				focus $focus
			}
		}
	}

	proc top {} { 
		upvar 1 self self
		return .msgContact-[join [lindex $self end] -]
	}

	if {[package vsatisfies [package present Tk] 8.4]} {
		proc status {status} {
			upvar 1 self self
			variable chats
			set top [top]
			set btn $top.edit.send
			$btn configure -image img:$status
			$self title
		}
	} else {
		proc status {status} {
			upvar 1 self self
			$self title
		}	
	}

	proc title {} { return Alicq }

	# On each alias change change window title as well
	proc alias {args} {
		upvar 1 self self
		$self title
	}

}
# End of single window chat dialog ```1

# Tabbed chat methods '''1
namespace eval TabbedChat {
	proc initialize {flag args} {
		upvar 1 self self
		set id [lindex $self end]
		set top .all
		set var [namespace current]::helpers::frameSelected
		if {![winfo exists $top]} {
			Toplevel $top AlicqUserWindow
			grid [frame $top.btn -class Tablist] -sticky new -row 0
			grid columnconfigure $top 0 -weight 1
			grid rowconfigure $top 1 -weight 1
			set $var ""
			trace variable $var w [nc helpers::SelectTab $top]
			bind $top <FocusIn> [namespace code { 
			     if {[winfo toplevel %W]!={%W}} continue
			}]
			set flag 1
		}
		set btn $top.btn.[helpers::win $id]

		radiobutton $btn -indicatoron 0\
			-text [$id alias] -value $id -variable $var
		pack $btn -side left -ipadx 2 -ipady 1
		frame [top] -class Tab
		if {$flag} { $btn invoke }
	}

	proc top {} {
		upvar 1 self self
		return .all.[helpers::win [lindex $self end]] 
	}

if {[package vsatisfies [package present Tk] 8.4]} {
	proc status {status} {
		upvar 1 self self
		set btn .all.btn.[helpers::win [lindex $self end]]
		$btn configure -image img:$status
	}
}	
	proc title {} { return AlicqChat }

	proc highlight {args} {
		upvar 1 self self
		helpers::highlight [lindex $self end] 
	}

	namespace eval helpers {
		bind Tab <Map> [namespace code { highlight $frameSelected }]
		bind Tab <Destroy> [nc DeleteTab %W]
		bind TabbedChat <<ToTab>> [nc ToTab %W %K]
		bind TabbedChat <<NextTab>> [nc NextTab %W 1]
		bind TabbedChat <<PrevTab>> [nc NextTab %W -1]

		proc win {id} { return w[join $id -] }
	}

}

proc TabbedChat::helpers::DeleteTab {win} {
	set parent [winfo parent $win].btn
	set btn $parent.[winfo name $win]
	if {![winfo exists $btn]} return
	set list [pack slaves $parent]
	if {[set pos [lsearch -exact $list $btn]]==-1} return
	set list [lreplace $list $pos $pos]
	if {![llength $list]} { 
		destroy [winfo toplevel $win]
	} else {
		destroy $btn
		if {![winfo exists [lindex $list $pos]]} { set pos end }
		[lindex $list $pos] invoke 
	}
}

# Select tab with keyboard
proc TabbedChat::helpers::ToTab {win key} {
	set slaves [winfo children [winfo toplevel $win].btn]
	set next [lindex $slaves [expr $key-1]]
	if {$next!=""} { $next invoke }
}

proc TabbedChat::helpers::NextTab {win step} {
	variable frameSelected
	set slaves [winfo children [winfo toplevel $win].btn]
	set me [winfo toplevel $win].btn.[win $frameSelected]
	set next [lindex $slaves [expr [lsearch -exact $slaves $me] + $step]]
	if {$next!=""} { $next invoke }
}

# Display tab selected with assotiated button
proc TabbedChat::helpers::SelectTab {top name args} {
	upvar 1 $name id
	set self [list chat $id]
	set win [[namespace parent]::top]
	if {![winfo exists $win]} return
	foreach x [grid slaves $top] { 
		if {$x==$win} return
		if {[winfo class $x]=="Tab"} {
			grid forget $x
			grid propagate [winfo parent $x] 0
		}
	}
	grid $win -sticky news -row 1
	$self title
	focus $win
}
# End of tabbed chat ```1

# Generic chat window '''1
namespace eval chat {
	variable chats
	method Chat initialize {args} {
		upvar 1 self self
		variable chats
		
		set ns [namespace parent]
		set top [$self top]
		set id [lindex $self end]
		foreach {chat in out} [${ns}::ChatBar $top.pw] break
		set chats($self) [list $in $out]
		grid $chat -sticky news
		grid [${ns}::EditBar $top.edit $id $in $out] -sticky we
		foreach x {row column} {grid ${x}configure $top 0 -weight 1}
		
		# Default bindings
		bind $top <FocusIn> [list ${ns}::DoFocus %W $top $out]
		foreach x [list $top $out] {
			bind $x <<Close>> [list destroy $top] 
		}
		bind $in <Destroy> [list $self destroy]

		# Create message formatter
		set Fm [list formatter [option get $in\
			messageStyle MessageStyle]]
		if {[info commands $Fm]==""} { set Fm {formatter default} }
		$Fm bind $in
		interp alias {} ${ns}::formatter::$id {} $Fm

		DropSite::register $out -dropcmd [nc DoDrop]\
			-droptypes {TREE_NODE copy}

		# Finally, Update status
		$self status [[lindex $self end] status]	

		# And build context menus
		ui::context-menu .incoming $self
	}

	method Chat in {} {
		upvar 1 self self
		variable chats
		lindex $chats($self) 0
	}	

	method Chat out {} {
		upvar 1 self self
		variable chats
		lindex $chats($self) 1
	}
	
	method Chat title {} {
		upvar 1 self self
		upvar 1 result suffix
		set id [lindex $self end]
		set status [string tolower [mc [$id status]]]
		wm title [winfo toplevel [$self top]]\
			"[$id alias] ($status, $id) - $suffix"
	}

	method Chat destroy {} {
		upvar 1 self self
		variable chats
		[lindex $self end] exclude [namespace parent]::active
		rename [namespace parent]::formatter::[lindex $self end] {}
		after idle [nc unset chats($self)]
	}

	method Chat::menu Cite {} {
		puts "Cite text"
	}

	proc DoDrop {txt tree args} {
		set node [lindex $args end]
		set uid [.tr itemcget $node -data]
		$txt image create end -padx 4 -image [.tr itemcget $node -image]
		$txt insert end [$uid alias] [list contact $uid]
	}

}
# End of generic chat ```1

# Default formatters '''1
namespace eval basic-formatter {
	method Formatter bind {win} {
		# Setup colors
		foreach x {incoming outgoing sent client server action
			authrequest} {
				set val [option get $win ${x}Color\
					[string totitle $x]Color]
			if {$val!=""} { $win tag configure $x -foreground $val }
		}
		foreach {x y} {alias ContactProperties action
		               AuthorizeFromDialog} {
			$win tag bind $x <Enter> { %W configure -cursor hand2 }
			$win tag bind $x <Leave> { %W configure -cursor {} }
			$win tag bind $x <1> [list [namespace parent]::$y %W]
		}
		$win tag configure action -underline 1
	}

	method Formatter time {time txt} {
		set fmt [option get $txt timeFormat TimeFormat]
		if {[clock seconds]-$time>12*3600} { 
			set fmt "[option get $txt dateFormat DateFormat] $fmt"
		}
		clock format $time -format $fmt
	}
}

# Default formatter 
namespace eval default-formatter {
	tags::new {formatter default}\
		[list [namespace current] 0.6 Formatter 0.5]

	proc bind {win} {
		# Underline given alias if needed
		set underline [option get $win underlineTag UnderlineTag]
		if {$underline!=""} { 
			$win tag configure $underline -underline yes
		}

		# Configure header options
		foreach x {background font relief} {
			set opt header[string totitle $x]
			set val [option get $win $opt [string totitle $opt]]
			if {$val!=""} { $win tag configure header -$x $val}
		}
		$win tag configure body\
			-foreground [option get $win bodyColor BodyColor]
		$win tag raise alias body
	}

	proc header {sender time win} {
		#list "\[$time\] " time [$sender alias] alias ":\n" ""
		list [$sender alias] alias "  " "" $time time "\n" ""
	}
}

# Sample Chat-style formatter
namespace eval chat-formatter {
	tags::new {formatter Chat} [list [namespace current] 0.6 Formatter 0.5]

	proc bind {win} {
		$win tag configure incoming -foreground red
		$win tag configure outgoing -foreground blue

		$win tag configure body\
			-foreground [option get $win bodyColor BodyColor]\
			-lmargin1 20 -lmargin2 20
		$win tag raise alias body
		$win tag configure delimiter -elide yes
	}

	proc header {sender time win} {
		list [$sender alias] alias " ($time)" time ": " ""
	}
}
# End of formatters ```1

proc DoFocus {from match to} { if {$match==$from} { focus $to } }

# Editbar contains buttons to control message edition and sending
proc EditBar {name self in out} {
	frame $name -class EditBar
	button $name.send -default active -command [nc SendMessage $self $out]
	bind $out <<Send>> "[list $name.send] invoke; break"
	grid $name.send -row 0 -column 0 -sticky wns
	# Check if id supports several message types
	set types [$self message-types]
	if {[llength $types]} {
		TypeButton $name.type $self $types $name.send $out
		grid $name.type	-row 0 -column 2 -sticky w
	}
	grid [button $name.close] -row 0 -column 10 -sticky e -padx 4
	$name.close configure -command [list event generate [winfo parent $name] <<Close>>]
	frame $name.aux -class AuxBar
	# If multicast mode, add contact aliases to citations
	button $name.aux.cite -command [nc CiteLast $in $out [$self is Group]]
	bind $out <<Cite>> [list $name.aux.cite invoke]
	grid $name.aux.cite -row 0 -column 1 -sticky wns
	# History button available only for unicast messages
	if {[$self is HistoryWriter]} {
		button $name.aux.history
		HistoryButton $name.aux.history $self $in [$self history-chunks]
		bind $out <<History>> [list $name.aux.history invoke]
		grid $name.aux.history -row 0 -column 2 -sticky wns
		checkbutton $name.dc -indicatoron 0 -text ""
		ui::tooltip::action $name.aux.history [mc "Show history"]\
			<<History>>
	}
	ui::tooltip::action $name.aux.cite [mc "Cite last message"] <<Cite>>
	ui::tooltip::action $name.close [mc "Close dialog"] <<Close>>
	ui::tooltip::action $name.send [mc "Send message"] <<Send>>
	grid $name.aux -sticky wns -row 0 -column 3 -ipadx 1 -padx 5
	foreach x {1 2 3} { grid columnconfigure $name $x -weight 1}
	set name
}

# Create menubutton for selection of message type from a list
# Selected type is stored in variable with UID-based name
proc TypeButton {name uid types send out} {
	set var [namespace current]::${uid}-type
	set values [map x $types {list $x [mc $x]}]
	ui::variant $name -variable $var -values $values 
	trace variable $var w [nc TypeSelected $uid $out]
	bind $name.menu <Destroy> [list + unset $var] 
	$name.menu invoke 0
	$send configure -command [join [list [$send cget -command]\
		[list $name.menu invoke 0]] "\n"]
	set name
}

proc TypeSelected {uid txt name args} {
	upvar 1 $name type
	set cmd [namespace current]::Type:$type
	if {[llength [info procs $cmd]]} { $cmd $uid $txt }
}

# Chatbar is panned window with text widgets for incoming and outgoing messages
# It uses either BWidget or native Tk panedwindow widget (depends on Tk version)
if {[package vsatisfies [package present Tk] 8.4]} {
	option add *Panedwindow*in.txt.state disabled
	option add *Panedwindow*Text.relief flat widgetDefault
	option add *Panedwindow.relief sunken widgetDefault
	option add *Panedwindow.sashPad 0 widgetDefault
	option add *Framein.height 200 widgetDefault
	option add *Frameout.height 100 widgetDefault
	option add *EditBar.AuxBar.cite.image button:cite widgetDefault

	image create bitmap button:cite -foreground brown -data {
		#define cite_width 16
		#define cite_height 16
		static unsigned char cite_bits[] = {
		   0,0,0x40,0,0xe0,0,0xf0,1,0xe0,3,0xc0,7,0x80,0x0f,0,0x1f,0,
		   0x1f,0x80,0x0f,0xc0,0x07,0xe0,3,0xf0,1,0xe0,0,0x40,0,0,0};
	}

	proc ChatBar {n} {
		panedwindow $n -orient vertical -showhandle 0
		foreach x {in out} {
			$n add [set $x [MFrame [frame $n.$x -class Frame$x] $x]]
		}
		list $n $in.txt $out.txt
	}

	proc TabbedChat::helpers::highlight {id {auto 0} {stage 1}} {
		variable flash
		set w [win $id]
		set vis [winfo viewable .all.$w]
		if {$vis} { 
			if {[info exists flash($id)]} {
				set stage 0
				after cancel flash($id)
				unset flash($id)
			} else return
		} else {
			if {!$auto && [info exists flash($id)]} return
			set flash($id) [after 300 [nc highlight $id\
				1 [expr $stage^1]]]
		}
		set image [expr {$stage?"message":[$id status]}]
		.all.btn.$w configure -image img:$image
	}
} else {
	option add *PanedWindow.f0*txt.state disabled
	option add *Frame.height 200 widgetDefault
	option add *EditBar.close.text x widgetDefault
	proc ChatBar {n} {
		PanedWindow $n -side right -activator line
		foreach x {in out} {set $x [MFrame [$n add -weight 1] $x]}
		list $n $in.txt $out.txt
	}

	proc TabbedChat::helpers::highlight {id} {
		set w [win $id]
		set btn .all.btn.$w
		set col [expr [winfo viewable .all.$w]?"foreground":"highlight"]
		set color [option get $btn $col Foreground]
		if {$color!=""} { $btn configure -foreground $color }
	}
}

proc MFrame {fr x} {
	variable tabbing
	set tag [expr {$tabbing?"TabbedChat":"SingleChat"}]
	bindtags $fr [concat MsgFrame Frame$x [bindtags $fr] SaveProportions]
	[ScrolledWindow $fr.sw] setwidget [text $fr.txt -wrap word -bd 1 -padx 8]
	bindtags $fr.txt [concat MsgText $tag Text$x [bindtags $fr.txt]]
	grid $fr.sw -sticky news
	grid propagate $fr 0
	foreach y {row column} {grid ${y}configure $fr 0 -weight 1}
	set fr
}

proc CiteLast {in out multi} {
	foreach {i1 i2} [$in tag prevrange incoming end] {
		set msg [$in get $i1 $i2]
		set prefix ""
		if {$multi} {
			set tags [$in tag names $i1]
			if {[set pos [lsearch $tags Contact:*]]!=-1} {
				set prefix [get [lindex $tags $pos] Alias]
			} 
		}
		foreach ln [lrange [split $msg "\n"] 1 end] {
			$out insert end "${prefix}> $ln\n" citation
		}
	}	
}

# Retrieve message from text widget, format according to it's type and
# send to recipient
proc SendMessage {id txt} {
	set msg [$txt get 1.0 "end -1 chars"]
	if {$msg=={}} return
	set var [namespace current]::${id}-type
	if {[info exists $var]} { set type [set $var] } else { set type text }
	set m $msg
	if {[llength [info commands Form$type]]} {set m [Form$type $msg]}
	$id send $type $m
	$txt delete 1.0 end
}

proc FormURL {msg} {
	set lines [split $msg "\n"]
	list [join [lrange $lines 1 end] "\n"] [lindex $lines 0] 
}

proc DisplayMessage {self sender time message {tags {}} {pos end} {see 1}} {
	set txt [[list chat $self] in]
	$txt configure -state normal
	$txt tag delete last
	set msg [list]
	foreach {part t} $message {lappend msg $part [concat $t body]}
	set header [list]

	foreach {part t} [formatter::$self header $sender\
				[formatter::$self time $time $txt] $txt] {
		lappend header $part [concat $t header]
	}

	set message [concat $header $msg]
	set msg [list]
	lappend tags time:$time from:$self
	foreach {part t} $message {lappend msg $part [concat $t $tags]}
	set msg [concat [lrange $msg 0 end-2] [list [string trimright\
			[lindex $msg end-1] "\n "] [lindex $msg end] "\n" ""\
			"\n" delimiter]]
	set from [$txt index $pos]		
	set fraction [lindex [$txt yview] end]
	eval [list $txt insert $from] $msg
	[list chat $self] highlight $from $pos
	if {$fraction > 0.99 || $see} {$txt see end}
	$txt configure -state disabled
	delayed [nc Raise [winfo toplevel $txt]]
	set txt
}

handler AddItem AddItem {class args} {
	set top .add-[string map {: -} $class]
	if {[winfo exists $top]} {
		raise $top; focus $top; return
	}
	foreach {domain type} [split $class :] break
	Toplevel $top AlicqAddItem
	wm title $top "Add new $type [string tolower $domain]"
	set meta [ref $class]::meta
	set cmd ""
	if {[info exists ${meta}::Identifier]} {
		set name [set ${meta}::Identifier]
		if {[info commands ::validate::$name]!=""} { 
			set cmd ::validate::$name
		}
	} else { set name Identifier }
	set var [namespace current]::$top
	grid [label $top.prompt] -sticky we -padx 8 -padx 2 -columnspan 2
	if {[info exists ${meta}::AddPrompt]} {
		$top.prompt configure -text [set ${meta}::AddPrompt]
	}
	grid [label $top.lbId -text [mc $name]]\
	     [entry $top.enId -validate key -vcmd\
	     		[nc ValidId $class $cmd $top.btn.add %P]]\
		-padx 8 -pady 4 -sticky new
	grid [label $top.lbAlias -text [mc Alias]] [entry $top.enAlias]\
		-padx 8 -pady 4 -sticky new
	grid [label $top.lbGroups -text [mc Groups]]\
		-columnspan 2 -padx 8 -pady 4 -sticky we
	grid [ScrolledWindow $top.sw]\
		-row 8 -columnspan 2 -padx 8 -pady 4 -sticky news
	$top.sw setwidget [Tree $top.trGroups -showlines 0 -deltay 20]
	grid [frame $top.btn -class ButtonBar] -row 11 -columnspan 2\
		-padx 8 -pady 4 -sticky wes
	button $top.btn.add -text [mc Add] -state disabled\
		-command [nc AddItemRun $class $top.enId $top.enAlias $var]
	button $top.btn.close -command [list destroy $top]
	grid $top.btn.add $top.btn.close -padx 4 -pady 2 -sticky e
	grid columnconfigure $top 1 -weight 1
	grid rowconfigure $top 8 -weight 1
	
	foreach {id alias groups} $args break
	if {[info exists id] && $id!=""} { $top.enId insert 0 $id }
	if {[info exists alias] && $alias!=""} { $top.enAlias insert 0 $alias }
	if {[info exists groups]} { foreach x $groups { set ${var}($x) 1} }

	GroupTree $top.trGroups $var
	set top
}
bind AlicqAddItem <Destroy> [nc unset %W]

# Prepare groups for recursive tree population
proc GroupTree {tree var} {
	array set hier [list]
	array set $var [list]
	foreach x [select Group:common] {
		set id [lindex [split $x :] end]
		if {$id=="other"} continue
		set groups [get $x Groups]
		if {$groups==""} { set groups ":" }
		foreach g $groups { lappend hier($g) $id }
	}
	variable nodecounter
	set nodecounter 0
	PopulateTree $tree hier $var root :
	unset hier nodecounter
}

# Recursively populate group tree, taking prepared information from data array
proc PopulateTree {tree data var parent node} {
	upvar 1 $data hier
	variable nodecounter
	if {![info exists hier($node)]} return
	foreach x $hier($node) {
		set name [set [ref Group:common:$x](Alias)]
		set node [$tree insert end $parent [incr nodecounter]\
			-data $x -open 1]
		$tree itemconfigure $node -window [checkbutton $tree.c$node\
			-text $name -variable ${var}($x)]
		PopulateTree $tree hier $var $node $x
	}
}

# Validate new object ID against its type, if given. Also check if object
# with same identifier already exists
proc ValidId {class cmd btn val} {
	if {$val=="" || [info exists [ref ${class}:${val}]]} {
		$btn configure -state disabled
		return 1
	} elseif {[set res [expr {($cmd=="")?1:[$cmd $val]}]]} {
		$btn configure -state normal
	}
	set res
}

# Get information for AddItem dialog and create new object
proc AddItemRun {class enId enAlias var} {
	set id [$enId get]
	set data(Alias) [$enAlias get]
	if {$data(Alias)==""} { unset data(Alias) }
	foreach {key val} [array get $var] { 
		if $val {lappend data(Groups) $key} 
	}
	destroy [winfo toplevel $enId]
	new $class:$id [array get data]
	switch $class {
		Contact:ICQ { 
			set oid [list ICQ $id] 
			set oclass ICQ
		}
		Group:common { 
			set oid [list group $id] 
			set oclass UserGroup
		}
	}
	if {[info exists oid]} {
		puts "Create object $oid $oclass"
		tags::new $oid [list $oclass 0.3] [array get data]
	}
}

proc HistoryButton {button id in chunks} {
	if {[llength $chunks]} {
		$button configure -command\
			[nc HistoryCmd $button $id $in $chunks]\
			-text "[mc History] (+[llength $chunks])"
	} else { $button configure -state disabled -text [mc "No history"] } 
}

proc HistoryCmd {button id in chunks} {
	$id history-chunk [lindex $chunks 0] [nc FillHistory $id]
	HistoryButton $button $id $in [lrange $chunks 1 end]
}

proc FillHistory {self type sender time msg tags chunk} {
	set txt [[list chat $self] in]
	set mark history:$chunk
	set mid [lindex $tags 0]
	if {[string match *:* $mid]&&[llength [$txt tag ranges $mid]]} return
	if {[lsearch [$txt mark names] $mark]==-1} {
		$txt mark set $mark 1.0
	} else { $txt mark gravity $mark right }
	DisplayMessage $self $sender $time $msg $tags $mark 0
	$txt mark gravity $mark left
	update idletasks
}

proc init {} {
	trace variable [namespace current]::style w [nc StyleChanged]
	set meta::style(values)\
		[map x [info commands [list formatter *]] { lindex $x end}]
	variable style
	if {[info exists style]} StyleChanged else { set style default }
}

proc StyleChanged {args} {
	variable style
	option add *Text.messageStyle $style
}

proc ContactProperties {txt} {
	set tags [$txt tag names current]
	if {[set pos [lsearch $tags from:*]]!=-1} {
		[string range [lindex $tags $pos] 5 end] menu::Info
	}
}

proc AuthorizeFromDialog {txt} {
	set tags [$txt tag names current]
	if {[set pos [lsearch $tags from:*]]!=-1} {
		set id [string range [lindex $tags $pos] 5 end]
	} else return
	if {[set pos [lsearch $tags action:*]]!=-1} {
		set action [lindex [split [lindex $tags $pos] :] end]
	} else return
	$id send authorization $action
}

# Generic messaging methods '''1
namespace eval messaging {
	include Contact [namespace current] 0.92
	proc incoming {type args} {
		upvar 1 self self
		if {[set [namespace parent]::popup]} {
			if {[popup $type 0]} {
				eval [list [namespace parent]::active::incoming\
					$type] $args
			}	
		} else { 
			$self include Pending 0.93
			eval [list $self pending $type] $args 
		}
	}

	proc popup {{type text} {flag 1}} {
		upvar 1 self self
		set ns [namespace parent]::active
		foreach cmd [info commands popup::$type] { 
			$cmd $self $flag
			$self include ${ns} 0.91
			return 1
		}
		return 0
	}

}

namespace eval messaging::popup {
	proc text {id flag} { 
		set ns [namespace parent [namespace parent]]
		upvar 0 ${ns}::tabbing t
		tags::new [list chat $id] [list Chat 0.5\
			${ns}::[expr {$t?"TabbedChat":"SingleChat"}] 0.4] $flag
	}

	foreach x {URL authrequest autorization} {
		interp alias {} [namespace current]::$x\
			     {} [namespace current]::text
	}
}
#```1

# Send contacts routines '''1
namespace eval messaging::popup::contacts {
	method ICQ::menu [list .11 Send Contacts] {} {
		upvar 1 self self
		$self popup outgoing-contacts
	}
	method Contact::menu [list .10 Send Message] {} {
		upvar 1 self self
		$self popup text
	}

}

proc messaging::popup::outgoing-contacts {id args} {
	set top .contacts-to-[join $id -]
	if {![catch {raise $top}]} return
	Toplevel $top AlicqContactsWindow
	wm title $top "Contacts for [$id alias]"
	set lbox [ListBox $top.list -deltay 20 -dropenabled yes\
		-droptypes {TREE_NODE} -dropcmd [nc contacts::Drop]]
	$lbox insert end zz -fill blue -text [mc "Drag contacts to send here"]
	grid $lbox -sticky news
	grid [frame $top.btn2 -class EditBar] -sticky wes -padx 4
	grid [button $top.btn2.send -command [nc contacts::Send $id $lbox]]\
			-sticky nws -row 1 -column 1
	grid [button $top.btn2.close -command [list destroy $top]]\
		-sticky nes -row 1 -column 5
	bind $top <<Send>> [list $top.btn2.send invoke]	
	bind $top <<Close>> [list $top.btn2.close invoke]	
	grid columnconfigure $top.btn2 4 -weight 1
	grid rowconfigure $top 0 -weight 1
	grid columnconfigure $top 0 -weight 1
}

proc messaging::popup::contacts::Send {recipient name} {
	foreach id [$name items 1 end] { 
		lappend res [lindex $id end] [$id alias] 
	}
	if {[info exists res]} { $recipient send contacts $res }
	destroy [winfo toplevel $name]
}

proc messaging::popup::contacts::Drop {widget src dest op datatype data} {
	set id [[winfo parent $src] itemcget $data -data]
	if {[$id is Group] || [$widget exists $id]} return
	$widget insert end $id -text [$id alias] -image img:[$id status]

}
# ```1

# Pending messages handling methods '''1
namespace eval pending {
	method Pending popup {args} {
		upvar 1 self self
		if {$args!=""} return
		variable pending
		foreach item $pending($self) {
			catch {eval [namespace parent]::active::incoming $item}
		}
		unset pending($self)
		$self no-pending
		$self exclude Pending
	}

	method Pending pending {args} {
		upvar 1 self self
		variable pending
		lappend pending($self) $args
	}
}
#```1

# Active chat messages handling '''1
namespace eval active {

	proc incoming {type args} {
		upvar 1 self self
		foreach cmd [info commands show::$type] {
			eval [list $cmd $self $self] $args incoming
		}
		return -code break
	}

	# Popup of active contact - raise window
	proc popup {args} {
		upvar 1 self self
		raise [[list chat $self] top]
		focus [[list chat $self] top]
		return -code break
	}

	proc send {type message} {
		upvar 1 self self
		upvar 1 result result
		set chat [list chat $self]
		if {[info exists result] && [info commands $chat]!=""} {
			foreach cmd [info commands show::$type] {
				$cmd my $self $message\
					[clock seconds] $result outgoing
			}
		} else { puts "msg not send" }
		return -code continue
	}

	proc acknowledged {msgid type} {
		upvar 1 self self
		set txt [[list chat $self] in]
		foreach {i1 i2} [$txt tag ranges $msgid] { 
			$txt tag add $type $i1 $i2
		}
	}
	
	proc new-status {status} {
		upvar 1 self self
		[list chat $self] status $status
	}

	# Monitor roster item alias change to update chat title if needed
	proc new-alias {alias} {
		upvar 1 self self
		after idle [list [list chat $self] alias $alias]
	}
}

namespace eval active::show {
	proc text {id in message time args} {
		[namespace parent [namespace parent]]::DisplayMessage\
			$in $id $time\
			[list $message {}] $args
	}

	proc URL {id in msg time args} {
		[namespace parent [namespace parent]]::DisplayMessage\
			$in $id $time\
			[list "[lindex $msg 0]\n" {} [lindex $msg 1] url]\
			$args
	}

	proc authrequest {id in msg time args} {
		foreach {uin nick fname lname email xxx reason} $msg break
		if {$id=="my"} {
			set msg [list $msg autorization]
		} else {
			set msg [list [lindex $msg end] {} "\n" {}\
			grant {action action:grant} "\t" {}\
			deny {action action:deny}]
		}
		[namespace parent [namespace parent]]::DisplayMessage\
		 	$in $id $time $msg $args
	}

	proc authorization {id in msg time args} {
		[namespace parent [namespace parent]]::DisplayMessage\
		 	$in $id $time [list "Authorization $msg"\
			authorization] $args
	}

	# Show incoming contacts in dialog
	proc contacts {id in msg time args} {
		set list [contacts::dialog $id]
		foreach {uin alias} $msg {
			puts "IN $uin $alias"
			if {[$list exists $uin]} continue
			$list insert end $uin\
				-text "$alias (ICQ $uin)"\
				-image img:offline -data $alias
		}
	}
}
#```1

# Display incoming contacts dialog '''1
namespace eval active::show::contacts {}
proc active::show::contacts::dialog {id} {
	set top .contacts-from-[join $id -]
	if {![catch {raise $top}]} { return $top.list }

	Toplevel $top AlicqContactsWindow
	wm title $top "Contacts from [$id alias]"

	grid [ListBox $top.list -dragenabled 1] -sticky news
	$top.list bindText <1> [list $top.list selection add]
	$top.list bindText <3> [list $top.list selection remove]
	grid [frame $top.btn -class ButtonBar] -sticky news -pady 2
	button $top.btn.add -command [nc Add $top.list] -text [mc Add]
	button $top.btn.close -command [list destroy $top]
	grid $top.btn.add x $top.btn.close -sticky e -padx 4
	bind $top <<Close>> [list $top.btn.close invoke]
	grid rowconfigure $top 0 -weight 1
	grid columnconfigure $top 0 -weight 1
	grid columnconfigure $top.btn 1 -weight 1
	return $top.list
}

proc active::show::contacts::Add {list} {
	foreach x [$list selection get] {
		set uid Contact:ICQ:$x
		if {![info exists [ref $uid]]} {
			new $uid [list Groups other\
				Alias [$list itemcget $x -data]]
			if {[info commands [list ICQ $x]]==""} {	
				tags::new [list ICQ $x] ICQ 0.4
			}
		} 
		$list delete $x
	}
	update idletasks
	if {![llength [$list items]]} { destroy [winfo toplevel $list]}
}

# Event bindings
bind MsgFrame <Destroy> {+
	option add *[string range %W 1 end].height [winfo height %W]
}

bind Textin <Map> [namespace code {
	%W tag bind incoming <3> {.incoming post %%X %%Y}
}]

if {$::tcl_platform(platform)=="windows"} {
	bind Textin <ButtonPress-1> { 
		if {![info exists [list focus-%W]] && [focus]!=[list %W]} {
			set [list focus-%W] [focus]
			focus %W
		}
	}

	bind Textin <KeyPress> { 
		if {[string is alnum "%A"]} {
			focus [set [list focus-%W]]
			event generate [set [list focus-%W]] <KeyPress>\
				-keycode %k -keysym %K
		}
	}

	bind Textin <FocusOut> {
		if {[info exists [list focus-%W]]} { unset [list focus-%W] }
	}
}

proc ClipboardCopy w {
	set own [selection own -displayof $w]
	if {$own!="" && [winfo toplevel $own]==[winfo toplevel $w]} {
		clipboard clear -displayof $w
		clipboard append -displayof $w [selection get]
	}
}
bind Textout <<Copy>> [nc ClipboardCopy %W]

bind Textout <Map> {
	%W tag configure citation -foreground brown
}

init


