# Binding for ICQ protocol library

package require icq 0.9.0

# Module metainformation '''1
namespace eval meta {
	set name "ICQ"
	set author "Ihar Viarheichyk <iverg@mail.ru>"
	set description "ICQ transport"
	set icon img:online
	
	# Guess default encoding
	proc Encoding {} {
		# Default value is system encoding
		set encoding [encoding system]
		# Use encodings beginnig with CP on the wire
		if {[string match cp* $encoding]} {return $encoding}
		# Trying find wire encoding by the list
		foreach {local wire} {{koi8-r koi8-u iso8859-5} cp1251} {
			if {[lsearch -exact $local $encoding]!=-1} {
				return $wire 
			}
		}
		# If previuos attempts failed, analyse LANG variable
		# (usefull for UTF-8 encoding
		if {[info exists ::env(LANG)]} {
			switch -regex $::env(LANG) {
				"^(be|ru|uk)_" 	{ set encoding "cp1251" }
			}
		}
		set encoding
	}

	array set uin {
		type digit description "ICQ UIN" weight .10
		property Global:ICQ:Info|Uin save change}
	array set alias {property Global:ICQ:Info|Alias save change
		weight .11}
	array set status {
		property Global:ICQ:Network|StartupStatus save change
		description "Status after startup" default offline
		type variant weight .12
		values {offline online occ dnd ffc away na invisible}
	}
	array set password { type password property Global:ICQ:Info|Password
		description "ICQ password" save change weight .101}
	array set encoding [list type variant save change\
		values [map x [encoding names] {set x [string index $x 0]:$x}]\
		desciption "Default encoding"\
		property Global:ICQ:Info|encoding weight .13]
	unset x	

}

# Network metainformation '''2
namespace eval network::meta {
	array set server {
		type fqdn default login.icq.com save change
		property Global:ICQ:Network|server weight .21
	}
	array set port {
		type port default 5190 save change weight .22
		property Global:ICQ:Network|port
	}
	array set proxy {
		type variant save change default "" weight .23
		description "Proxy to use" empty "No proxy"
		valuescript {map x [select Proxy] { lindex [split $x :] end}}
	}
	array set reconnect {
		type boolean save change property Global:ICQ:Network|reconnect
		description "Automatically re-connect after disconnect"
		default 1 weight .24
	}
	array set ping {
		type integer save change property Global:ICQ:Network|ping
		description "Ping interval, seconds" default 0 weight .25
	}
	array set keepalive {
		type integer save change property Global:ICQ:Network|keepalive
		description "Keepalive interval, seconds" default 0 weight .26
	}
}

# Roster metainformation '''2
namespace eval roster::meta {
	array set use { 
		type boolean default 1 property Global:ICQ:Roster|use
		save change weight .31
		description "Use server-side roster"
	}
	array set modify {
		type boolean default 1 property Global:ICQ:Roster|modify
		save change weight .32
		description "Save Changes in contact list in server-side roster"
	}
	array set updated {
		type cache property Global:ICQ:Roster|updated
		save exit default 0
	}
	array set count {
		type cache property Global:ICQ:Roster|count
		save change default 0
	}
}	
#```1

# ICQ-specific properties for ICQ contacts '''1
namespace eval [ref Contact:ICQ]::meta {
	set Identifier UIN
	set AddPrompt { You should specify numerical UIN of ICQ contact. 
	}
	array set Status  {
		type {cache variant} default offline 
		values {offline online occ dnd ffc away na invisible}
	}
	array set mobile 	{save change}
	array set auth		{type {cache boolean} default 0 save change}
	array set rid		{type integer save change}
	array set visible	{type boolean default 0 save change }
	array set invisible	{type boolean default 0 save change }
	array set encoding	{ save change }
}

# ICQ contact main menu items
namespace eval menu {
	method ICQ::menu [list .20 Autorization Grant] {} {
		upvar 1 self self
		$self send authorization grant
	}

	method ICQ::menu [list .21 Autorization Deny] {} {
		upvar 1 self self
		$self send authorization deny
	}

	# Update contact encoding from convext menu
	method ICQ::menu Encoding {args} {
		upvar 1 self self
		set ref [$self property encoding]
		if {$args==""} {
			if {[info exists $ref]} { 
				set enc [set $ref]
			} else {set enc [set [namespace parent]::encoding] }	
			return [list $enc [map x [encoding names] {
					set x [string index $x 0]:$x}]]
		}
		set $ref [lindex $args 0]
		upvar 0 [namespace parent]::icq icq
		if {![info exists icq]} return
		$icq encoding [lindex $args 0] [lindex $self end]
	}
}

set ref [ref Contact:ICQ]
array set ${ref}::meta::encoding [array get meta::encoding]

# Sync contact default encoding with global encoding
upvar #0  ${ref}::meta::encoding(default) [namespace current]::encoding

namespace eval [ref Group:common]::meta {
	array set rid	{type integer save change}
}

# Data validation functions '''1
namespace eval ::validate {
	proc port {val} {
		expr { [string is integer $val] && $val>=0 && $val<=65535 }
	}
	proc UIN {val} { string is digit $val }
}
# ```1

variable icq
variable recreate 0
variable itemCount 0

# Process ICQ events '''1
proc ICQEvent {event args} {
	if {[info commands hooks::$event]!=""} {
		eval [list hooks::$event] $args
	} elseif {[llength [info commands icq:$event]]} {
		eval icq:$event $args
	} else { eval Event [list $event] $args}
}

namespace eval hooks {
	proc Log {tags str} { Event Log [concat $tags ICQ] $str }

	proc Status {uin newstatus} { [list ICQ $uin] new-status $newstatus }
	
	proc IP {uin ip} {
		if {$ip!="0.0.0.0"} { set [ref Contact:ICQ:$uin](IP) $ip }
	}

	proc LanDetails {uin type ip port cookie version} {
		[list ICQ $uin] dc-info $type $ip $port $cookie $version
	}

	proc Client {uin args} {
		set ref [ref Contact:ICQ:$uin]
		set ${ref}(Client) $args
		[list ICQ $uin] client-info $args
	}

	proc ACK {type uin msgid} {
		Event Log info "Ack $type from $uin on $msgid"
		[list ICQ $uin] acknowledged $msgid $type
	}

	proc MyIP {local remote} {
		Event Log info "Local IP $local, remote IP $remote"
	}

	proc Capabilities {uin caps} {
		set [ref Contact:ICQ:$uin](Capabilities) $caps
		Event Log info "Capabilities $uin: $caps"
	}

	# Process personal status
	interp alias {} [namespace current]::MyStatus {} my new-status

	# Translate ICQ message to alicq messages
	proc Incoming {type sender time message} {
		set id [clock seconds]:[clock clicks] 
		set oid [list ICQ $sender]
		set message [string trimright $message "\0"]
		if {$type=="web"} {
			Event Incoming text Contact:ICQ:0\
				$time "$sender:\n$message" $id
		} else {
			if {$type=="auto"} { 
				set type "text"
				set [ref Contact:ICQ:$sender](description)\
					$message
			} elseif {$type=="authorization" &&
			          $message=="granted"} {
				set [ref Contact:ICQ:$sender](auth) 0
				$oid exclude Unauthorized
			
			}
			$oid incoming $type $message $time $id
		}
	}

	proc Registered {uin password} {
		set [namespace parent]::recreate 1
		set [namespace parent]::uin $uin
		set [namespace parent]::password $password
		[namespace parent]::SetStatus online
	}

	proc RegistrationRefused {} {
		Event Error registration\
			"ICQ server refused registration of new UIN"
	}

	proc icq:Missed {uin reason channel} {
		set uid Contact:ICQ:$uin
		Event Log {error missed} "Missed message from $uin: $reason"
		set auto [mc "Message from this contact was discarded by server."]
		if {[string is integer $reason]} {
			append auto "\n" [mc "The description of the reason is not available, please contact autor of Alicq and report error code."] \
				"\n" [mc "Error code"] " " $reason
		} else { append auto "\n" [mc "Reason:"] " " [mc $reason] }
		[list ICQ $uin] incoming text $auto [clock seconds] 0
	}
}
# End of ICQ events ```1

# Process contact statuses
method ICQ new-status {status} {
	upvar 1 self self
	set [ref Contact:[join $self :]](Status) $status
}

method ICQ status {} {
	upvar 1 self self
	set var [ref Contact:ICQ:[lindex $self end]](Status)
	if {[info exists $var]} { return -code return [set $var] }
	return offline
}

method ICQ message-types {} { list text URL authrequest }

method ICQ hint {} {
	upvar 1 self self
	upvar 0 [ref Contact:[join $self :]] info
	
	if {[info exists info(Client)]} {
		variable encoding
		foreach {ver client unicode} $info(Client) break
		if {[info exists info(encoding)]} {
			set enc $info(encoding)
		} else { set enc $encoding }
		lappend t "$client, ICQv${ver}"\
			[expr ${unicode}?"Unicode":"Encoding: $enc"]
	}
	if {[info exists info(client:name)]} { lappend t $info(client:name) }
	if {[info exists info(description)]} {
		lappend t "($info(description))"
	}
	if {[info exists info(IP)]} { lappend t "IP: $info(IP)" }
	if {![info exists t]} { return -code continue }
	join $t "\n"
}

# On contact delete, remove it from list of all contacts as well
method ICQ destroy {} {
	upvar 1 self self
	variable icq
	if {[info exists icq]} { $icq contacts all delete [lindex $self end] }
}

# Transform method to old-style event
method ICQ incoming {type message time id} {
	upvar 1 self self
	Event Incoming $type Contact:[join $self :] $time $message $id
}

# Send ICQ message
method ICQ send {type message} {
	upvar 1 self self
	variable icq
	if {[info exists icq]} {
		# Check if message already sent e.g. by DC
		upvar 1 result result
		if {[info exists result]} { return -code continue }
		return [$icq send $type [lindex $self end] $message]
	} else { return -code continue }
}

proc icq:Roster:Items {items time} {
	variable itemCount
	foreach x $items {
		foreach {type id gid name aux} $x break
		if {!$gid && $type=="group"} {
			incr itemCount
			continue
		}
		if {![roster::item $type $name $id $gid $aux]} { 
			incr itemCount 
		}
	}
	if {$time} { 
		set roster::updated $time
		set roster::count $itemCount
		set itemCount 0
		roster::sync
	}
}

proc icq:Roster:Update {status} {
	roster::result $status
}

proc EstimatedRoster {} {
	expr { [llength [select {Contact:ICQ Group} {[info exists rid]}]]+
	       ${roster::count} }
}


# Handle ICQ Server-side roster '''1
namespace eval roster {
	namespace eval account {
		include ICQ-Account [namespace current] 0.2

		proc new-status {status} {
			if {$status != "offline"} return
			# Disable server-side specific features

			exclude ICQ-Roster ICQ-Roster
			exclude UserGroup [namespace parent]::monitor

			foreach x {gid rid mapping} {
				set var [namespace parent]::items::$x
				if {[info exists $var]} { unset $var }
			}
		}
	}

	# New ICQ object is added, include to SSI
	namespace eval monitor {
		proc initialize {args} {
			upvar 1 self self
			[namespace parent]::just-add $self 0\
				[[namespace parent]::rid] [$self alias]
		}
	}

	# Add ICQ contact to server-side roster if it is added to one of 
	# roster groups
	method ICQ-Roster add {id} {
		upvar 1 self self
		if {![$id is ICQ]} { return -code continue }
		add-item $id [lindex [lindex $items::rid($self) 0] end]
		return -code continue
	}

	proc add-item {id gid} {
		# Check we are not in this group yet
		if {[info exists items::rid($id)]} {
			foreach item $items::rid($id) {
				if {[lindex $item end] == $gid} return
			}
		}
		# Send future authorization to contact
		if {![$id is ICQ-Roster]} { $id send futureauth {} }
		just-add $id [rid] $gid [lindex $id end]
	}

	proc rid {} {
		# Generate new rid in range 2000-32000
		set rid 2000
		while {[info exists items::mapping($rid)]} {
			set rid [expr {int(rand()*30000.0+2001)}]
			# TODO: infinite loop is theoretically possible
		}
		set rid
	}

	proc just-add {id rid gid name {run 1}} {
		upvar 0 [namespace parent]::icq icq
		variable rosterop
		# Add to roster
		lappend rosterop [list [nc added-ok $id $rid $gid]\
			[nc added-error $id $rid $gid]]
		set r [list [type $id] $rid $gid $name [aux $id]]
		if {$run} { $icq roster add $r } else { return $r }
	}

	proc added-ok {id rid gid} {
		my log info "$id added to SSI"
		if {![$id is ICQ-Roster]} { $id include ICQ-Roster }
		lappend items::rid($id) [list $rid $gid]
		set items::mapping($rid) $id
		# If group was added, check if any contacts in it 
		# should to be added
		if {[$id is UserGroup]} {
			set [$id property rid] $gid
			foreach x [tags::select ICQ] {
				if {[$x is ICQ-Roster]} continue
				foreach y [$x parents] {
					if {$y==$id} { 
						add-item $x $gid
						break
					}
				}
			}
		}
	}

	proc added-error {id rid gid reason} {
		# In case of Error 14 try to add as nonauthorized
		if {$reason == 14} {
			# Check if we tried to add it before
			if {![$id is Unauthorized]} {
				$id include Unauthorized
				# Try to add to roster again
				just-add $id $rid $gid [lindex $id end]
			}
		}
	}

	# ICQ contact deleted, remove from SSI
	method ICQ-Roster destroy {} {
		upvar 1 self self
		upvar 0 [namespace parent]::icq icq
		variable rosterop
		set type [type $self]
		if {$type=="contact"} {
			set name [lindex $self end]
		} else {
			set name [$self alias]
		}
		foreach item $items::rid($self) {
			foreach {rid gid} $item break
			lappend rosterop [list [nc deleted-ok $self $rid -1] {}]
			lappend q delete [list $type $rid $gid $name\
				[aux $self]]
		}
		eval [list $icq roster] $q
	}

	# If item was deleted from SSI without errors, remove local mappings
	proc deleted-ok {id rid p} {
		if {$p!=-1} {
			set items::rid($id) [lreplace $items::rid($id) $p $p]
		} else { 
			if {[info exists items::rid($id)]} {
				unset items::rid($id) 
			}	
		}
		unset items::mapping($rid)
	}

	# ICQ contact or group alias change
	method ICQ-Roster new-alias {args} {
		upvar 1 self self
		upvar 0 [namespace parent]::icq icq
		variable rosterop
		set type [type $self]
		if {$type=="contact"} {
			set name [lindex $self end]
		} else {
			set name [$self alias]
		}
		foreach item $items::rid($self) {
			foreach {rid gid} $item break
			lappend rosterop [list {} {}]
			lappend q update [list $type $rid $gid $name\
				[aux $self]]
		}
		eval [list $icq roster] $q
	}

	# Any contact removed from roster group.
	# This method monitors group movements
	method ICQ-Roster remove {id} {
		# Only ICQ roster contacts can be removed from SSI
		if {![$id is ICQ] || ![$id is ICQ-Roster]} { 
			return -code continue 
		}
		upvar 1 self self
		# Check if out group is primary for contact we are going to
		# remove
		upvar 0 [namespace parent]::icq icq
		variable rosterop
		set group [lindex [lindex $items::rid($self) 0] end]
		set pos 0
		foreach item $items::rid($id) {
			foreach {rid gid} $item break
			if {$gid == $group} {
				set uin [lindex $id end]
				lappend rosterop [list\
					[nc deleted-ok $id $rid $pos] {}]
				$icq roster delete [list contact $rid $gid\
					$uin [aux $id]]
			}
			incr pos
		}
	}
	
	# Intercept authorization granted messages
	method ICQ-Roster incoming {type message args} {
		upvar 1 self self
		if {($type=="authorization" || $type == "futureauth") 
		    && [$self is Unauthorized]} {
		    	$self exclude Unauthorized
		}
		return -code continue
	}

	proc aux {id} {
		if {[lindex $id 0]=="group"} { return [list] }
		# For contacts aux info is alias, group and auth flag
		set res [list Alias [$id alias]]
		if { [$id is Unauthorized] } { lappend res auth "" }
		set res
	}

	proc type {id} { string map {ICQ contact} [lindex $id 0] }

	# This procedure is invoked after SSI was processed to add groups and
	# contact added when offline
	proc sync {} {
		foreach x [tags::select UserGroup] {
			if {[$x is ICQ-Roster]} continue
			lappend q add [just-add $x 0 [rid] [$x alias] 0]
		}
		foreach x [tags::select ICQ] {
			if {[$x is ICQ-Roster]} continue
			$x include Unauthorized
			foreach g [$x parents] {
				if {[$g is UserGroup] && [$g is ICQ-Roster]} {
					set gid [lindex [lindex\
						$items::rid($g) 0] end]
					lappend q add [just-add $x [rid] $gid\
						[lindex $x end] 0]
				}
			}
		}
		if {[info exists q]} {
			upvar 0 [namespace parent]::icq icq
			#eval [list $icq roster] $q
		}
		# Intercept new objects initialization
		include UserGroup [namespace current]::monitor 0.95
	}

	# Handle status of SSI operation
	proc result {status} {
		variable rosterop
		Event Log debug "Contact updated with status $status"
		if {[info exists rosterop]} {
			foreach {onok onerror} [lindex $rosterop 0] break
			set rosterop [lrange $rosterop 1 end]
			if {$status} {
				if {$onerror!=""} { eval $onerror $status }
				my log error "Error updating SSI: $status"
			} else {
				if {$onok!=""} { eval $onok }
			}
		}
		update idletasks
	}
	
	# Handle incoming item
	proc item {type name rid gid aux} {
		if {[info commands items::$type]==""} { return 0 }
		set o [items::$type $name $rid $gid $aux]
		if {![info exists items::rid($o)]} { 
			$o include ICQ-Roster 0.8 
		}
		lappend items::rid($o) [list $rid $gid] 
		return 1
	}
}

namespace eval roster::items {
	proc contact {name rid gid aux} {
		variable mapping
		set id [list ICQ $name]

		array set opt $aux

		# Map contact group
		if {[info exists mapping($gid)] &&
		    [info commands $mapping($gid)]!=""} {
		    	set group $mapping($gid)
		}

		if {[info commands $id]!=""} {
			if {[info exists opt(Alias)] &&
			    $opt(Alias)!=[$id alias]} { 
			    	$id new-alias $opt(Alias)
			}
			if {[info exists group]} {
				foreach x [$id parents] {
					if {$x == $group} {
						unset group
						break
					}
				}
				if {[info exists group]} {
					after idle [list $group add $id]
				}
			}
		} else { 
			if {[info exists group]} {
				set opt(Groups) [list [lindex $group end]]
			}
			tags::new $id {ICQ 0.3} [array get opt] 
		}

		if {[info exists opt(auth)]} { $id include Unauthorized }
		set mapping($rid) $id
		set id
	}

	proc group {name rid gid args} {
		variable mapping
		# Search group by saved rid
		foreach x [tags::select UserGroup] {
			set var [$x property rid]
			if {[info exists $var] && [set $var]==$gid} {
				set id $x
				break
			}
		}
		# Map by rid if not found
		if {![info exists id]} { set id [list group $gid] }

		if {[info commands $id]!=""} {
			if {$name != [$id alias]} { $id new-alias $name }
		} else { tags::new $id { UserGroup 0.3} [list Alias $name] }
		set mapping($gid) $id

		set var [$id property rid]
		if {![info exists $var] || [set $var]!=$gid} { set $var $gid }
		set id
	}
}
# End of SSI handling ```1

# Fake icq command is used when no UIN is specified.
# It always returns empty string as result and sends Error:authorization
# when trying to change status
proc fake_icq {sub args} {
	if {$sub=="status" && [lindex $args 0]!="offline"} {
		Event Error:auth "UIN is not specified or non-numeric"
	}
	list
}

proc CreateConnection {} {
	set icq [icq::icq\
		-capabilities {ICQ RELAY UTF8}\
		-event [namespace current]::ICQEvent]
	if {[string is true $roster::use]} {
		if {$roster::updated>0} {
			Event Log info "Estimated roster: [EstimatedRoster], $roster::updated"
			set roster [list $roster::updated [EstimatedRoster]]
		} else {set roster 1}
	} else {set roster 0}
	$icq configure -roster $roster

	variable encoding
	$icq encoding $encoding

	# Create contact lists
	foreach id [tags::select ICQ] {
		set cuin [lindex $id end]
		lappend all $cuin
		set ref [$id property]
		# Generate visible and invisible lists
		foreach x {visible invisible} {
			if {[info exists ${ref}($x)] && [set ${ref}($x)]} {
				lappend $x $cuin
			}
		}
		# Set encoding if speficied
		if {[info exists ${ref}(encoding)]} {
			$icq encoding [set ${ref}(encoding)] $cuin
		}
	}
	foreach x {all visible invisible} {
		if {[info exists $x]} { $icq contacts $x add [set $x] }
	}
	
	variable dcserver
	if {[info exists dcserver]} { $icq configure -dcinfo [$dcserver info] }
	set icq
}

# Register new ICQ UIN. New icq command created for this purpose
handler ICQ:Register registrator {password} {
	# Cancel previous registration
	Event ICQ:CancelRegistration
	icq::icq RegistratorICQ
	foreach x [info variables network::*] {
		RegistratorICQ configure -[namespace tail $x] [set $x]
	}
	RegistratorICQ register $password
}

# Cancel ICQ UIN registration. Only one regustration command at the time
# allowed, thus no task identifiers required
handler ICQ:CancelRegistration registration_canceler {args} {
	if {[info commands RegistratorICQ]} {
		RegistratorICQ delete
	}
}

# Track changes of contact's encoding 
# TODO: move DC handling to menu
proc trackencoding {uin ref field args} {
	variable icq
	if {![info exists icq]} return
	upvar 1 ${ref}($field) val
	$icq encoding $val $uin
	if {[info exists ${ref}(dc)]} { 
		[set ${ref}(dc)] configure -encoding $val
	}
}

# Track changes of global encoding
proc GlobalEncoding {ref args} {
	variable icq
	if {![info exists icq]} return
	upvar 1 $ref val
	$icq encoding $val
}

proc Reconfigure {name field op} {
	variable icq
	upvar 1 $name var

	if {[catch { $icq configure -[namespace tail $name] $var} r]} {
		Event Log error "Cant configure $name: $r"
	}
}

proc init {} {
	after idle [nc configure]
}

proc configure {args} {
	variable icq
	# If encodin was not set, try to guess default one
	variable encoding
	if {![info exists encoding]} {
		set encoding [meta::Encoding]
		set meta::encoding(default) $encoding
		# We do not need Encoding proc any more, drop it
		rename meta::Encoding {}
	}
	trace variable [namespace current]::encoding w [nc GlobalEncoding]

	foreach x {alias status} { variable $x }

	if {[info exists alias]} { set [ref Me](Alias) $alias } else {
		trace variable [ref Me](Alias) w [nc SetAlias]
	}
	set icq [CreateConnection]

	foreach x {uin password} {
		Reconfigure [namespace current]::$x "" ""
		trace variable [namespace current]::$x w [nc Reconfigure]
	}

	foreach x [info vars network::*] {
		Reconfigure $x "" ""
		trace variable $x w [nc Reconfigure]
	}


	if {$status!="offline"} { SetStatus $status }
}

# Basic methods available for ICQ contacts
# New ICQ contact added
method ICQ initialize {properties args} {
	upvar 1 self self
	variable icq
	if {![info exists icq]} return 
	
	# Add to list to monitor status
	$icq contacts all add [lindex $self end]
	
	# Check if contact has alias, request user info if has not
	array set opts $properties
	if {![info exists opts(Alias)]} {
		$icq info [lindex $self end] [nc DefaultAlias $self]
	} else {
		#TODO: add to alias pending queue
	}
}

proc SetStatus {setstatus} {
	variable icq
	$icq status $setstatus
}

method Me::menu {Change "ICQ Password"} {} {
	package require AuthDialog
	PasswordDialog .icqpwd -title "Chance ICQ password"\
		-info "Enter new password\nfor your ICQ account"\
		-ok [nc ChangePassword]
}

proc ChangePassword {password} {
	variable icq
	$icq password $password $ref
}

proc onOnline {} {
	# TODO: request aliases for contacts added when offline
}

proc DefaultAlias {id info} {
	array set opt $info
	if {[info exists opt(Nick)]} { 
		$id new-alias $opt(Nick)
	}
}

# Sync icq alias when global alias changes
proc SetAlias {name idx args} {
	upvar 1 ${name}($idx) val
	variable alias $val
}

proc GetAlias {args} {
	foreach x {icq uin} { variable $x }
	if {[set [ref Me](Status)]=="offline"} return
	trace vdelete [ref Me](Status) w [nc GetAlias]
	set ref [expr {[clock seconds] & 0xFF}]
	#hook Info [nc NewContactInfo Me $ref]
	#$icq info $uin $ref
}

# ICQ account methods '''1
namespace eval ICQ-Account {
	my include ICQ-Account 0.3

	method ICQ-Account status {} { 
		return -code return [set [ref Me](Status)] 
	}

	method ICQ-Account new-status {new} {
		if {$new=="offline"} {
			foreach x [tags::select ICQ] { 
				if {[$x status]!="offline"} {
					$x new-status offline
				} 
			}
		}
		set [ref Me](Status) $new
	}

	method ICQ-Account set-status {new} {
		[namespace parent]::SetStatus $new 
		if {$new=="offline"} { 
			upvar 1 self self
			$self new-status offline 
		}
	}

	method ICQ-Account icq-command {} {
		variable icq
		set var [namespace parent]::icq
		if {[info exists $var]} { return -code return [set $var] }
	}
}
# End of ICQ account methods ```1

# ICQ special groups '''1
namespace eval groups {
	proc parents {type} { return -code return [list {tree special}] }
	
	proc alias {type} { return -code return "ICQ [string totitle $type]"}

	proc add {type id} {
		set [$id property $type] 1 
		upvar 0 [namespace parent]::icq icq
		if {[info exists icq]} {
			$icq contacts $type add [lindex $id end]
		}
	}

	proc remove {type id} { 
		set [$id property $type] 0 
		upvar 0 [namespace parent]::icq icq
		if {[info exists icq]} {
			$icq contacts $type delete [lindex $id end]
		}
	}
}

namespace eval groups::Unauthorized {
	tags::new {special Unauthorized} [list [namespace current] 0.3\
		Group 0.5]

	proc parents {} { return -code return [list {tree special}] }

	proc hint {} { return -code return "Contacts not authorized by you" }
}

namespace eval groups::invisible {
	tags::new {special invisible} [list [namespace current] 0.3\
		Group 0.5]

	# Make aliases to common methods
	foreach x {parents alias add remove} {
		interp alias {} [namespace current]::$x {}\
			[namespace parent]::$x invisible
	}

	proc hint {} { return -code return "Members of this group will not see your status in normal mode" }
}

namespace eval groups::visible {
	tags::new {special visible} [list [namespace current] 0.3\
		Group 0.5]

	# Make aliases to common methods
	foreach x {parents alias add remove} {
		interp alias {} [namespace current]::$x {}\
			[namespace parent]::$x visible
	}

	proc hint {} { return -code return "Only members of this group can see your status in invisible mode" }
}

# End of special groups ```1

# Add contact to virtual groups if needed
method ICQ parents {} {
	upvar 1 result result
	upvar 1 self self
	foreach x {visible invisible} {
		set ref [$self property $x]
		if {[info exists $ref] && [set $ref]} {
			lappend result [list special $x]
		}
	}
	if {[info exists result]} { return $result }
}

method Me::menu {Add Contact} {} { Event AddItem Contact:ICQ }
include ICQ Contact 0.5

init
