# Sample IRC module for Alicq IM

namespace eval meta {
	proc unload {} { {tree IRC} delete }
}

namespace eval server-list {
	tags::new {tree IRC} [list [namespace current] 0.5 RosterItem 0.7]

	proc parents {} { return -code return [list {tree root}] }

	proc hint {} { 
		return -code return "IRC servers"
	}

	proc delete {} {
		foreach x [tags::select IRC-Server] { $x delete }
	}

	namespace eval menu {
		proc {Add IRC server} {} {
			puts "Not implemented yet"
		}
	}
}

namespace eval server {
	include IRC-Server Group 0.5

	method IRC-Server parents {} { list {tree IRC} }

	method IRC-Server image {} {
		upvar 1 self self
		return img:[status]
	}

	method IRC-Server initialize {name {port 6667}} {
		variable servers
		upvar 1 self self
		set servers($self) [list $name $port]
	}

	method IRC-Server delete {} {
		upvar 1 self self
		variable servers
		# Disconnect
		disconnect
		unset servers($self)
		# Delete related channels
		foreach x [tags::select IRC-Channel] { 
			if {[lindex $x 0]==$self} { $x delete }
		}
		puts "irc-server $self deleted"
	}

	method IRC-Server new-status {status} {
		upvar 1 self self
		if {$status=="online"} {
			tags::new [list $self #tcl] {IRC-Channel 0.5}
		}
	}

	# Get server status
	method IRC-Server status {} {
		upvar 1 self self
		variable sockets
		return -code return [expr\
			{[info exists sockets($self)]?"online":"offline"}]
	}

	# Map server status change to either connect or disconnect
	method IRC-Server set-status {status} {
		upvar 1 self self
		if {$status=="offline"} disconnect else connect 
	}

	# Handle online status request (connect)
	method IRC-Server connect {} {
		upvar 1 self self
		variable servers
		variable sockets
		if {[info exists sockets($self)]} return
		foreach {host port} $servers($self) break
		set sockets($self) [socket -async $host $port]
		fconfigure $sockets($self) -blocking no -translation crlf\
			-buffering none
		fileevent $sockets($self) writable\
			[nc network::Init $sockets($self) $self]
	}

	# Handle offline status request (disconnect)
	method IRC-Server disconnect {} {
		upvar 1 self self
		variable sockets
		if {[info exists sockets($self)]} { 
			puts "disconnecting $self"
			close $sockets($self)
			unset sockets($self)
			$self new-status offline
		}
	}

	method IRC-Server command {str} {
		upvar 1 self self
		variable sockets
		if {[info exists sockets($self)]} { puts $sockets($self) $str }
	}
}

namespace eval channel {
	include IRC-Channel RosterItem 0.8

	method IRC-Channel initialize {args} {
		upvar 1 self self
		set server [lindex $self 0]
		if {[$server status]=="online"} join
	}

	method IRC-Channel join {} {
		upvar 1 self self
		variable status
		if {[status]!="offline"} return
		foreach {server channel} $self break
		$server command "join $channel"
	}

	# IRC channel parent is it's server
	method IRC-Channel parents {} {
		upvar 1 self self
		list [lindex $self 0]
	}

	# Menu method for join
	method IRC-Channel::menu Join {} {
		upvar 1 self self
		join
	}

	method IRC-Channel new-status {new} {
		upvar 1 self self
		variable status
		set status($self) $new
		if {$new=="online"} {
			[namespace parent]::ui::chat $self
		}
	}

	method IRC-Channel status {} {
		upvar 1 self self
		variable status
		if {[info exists status($self)]} {
			return $status($self)
		} else { return "offline" }
	}

	method IRC-Channel image {} { 
		upvar 1 self self
		set img img:[status]
		puts "image for $self is $img"
		return -code return $img
	}

	method IRC-Channel send {text} {
		upvar 1 self self
		foreach {server channel} $self break
		$server command "privmsg $channel :$text"
	}
}

namespace eval server::network {
	variable host [exec hostname]

	proc nick {} { return "igor2" }
	proc full-name {} { return "I V" }
}

proc server::network::Init {socket self} {
	variable host
	puts $socket "user $::tcl_platform(user) $host $host [full-name]"
	puts $socket "nick [nick]"
	fileevent $socket writable {}
	fileevent $socket readable [nc Read $socket $self]
}

proc server::network::Read {socket self} {
	set count [gets $socket str]
	if {$count==-1} {
		if {[eof $socket]} {
			puts "Close connection"
			close $socket
		}	
		return
	}
	if {[regexp {(:([^ ]+) )?([0-9a-zA-Z]+) (.*)$} $str _ F from cmd p]} {
		if {[set idx [string first " :" $p]]} {
			set params [string range $p 0 [expr $idx-1]]
			lappend params [string range $p [expr $idx+2] end]
		} else { set params $p }
		set cmd [string toupper $cmd]
		if {[info exists mapping($cmd)]} { set cmd $mapping($cmd) }
		puts "check $cmd: $params"
		set cmd [namespace parent]::hooks::$cmd
		if {[info commands $cmd]==$cmd} {
			eval [list $cmd] $params
		}
	}
}

namespace eval server::hooks {
	proc 005 {args} {
		upvar 1 self self
		puts "$self: Seems like we are connected"
		$self new-status online
	}

	proc 700 {nick encoding args} {
		upvar 1 socket socket
		puts "Set socket encoding to $encoding"
		fconfigure $socket -encoding [string tolower $encoding]
	}

	proc JOIN {channel args} {
		upvar 1 self self
		[list $self $channel] new-status online
	}

	proc 353 {nick _ channel users} {
		puts "Users on channel $channel: $users"
	}

	proc PRIVMSG {to message} {
		upvar 1 self self
		upvar 1 from from
		[list $self $to] incoming $from $message
	}

	proc PING {origin} {
		upvar 1 socket socket
		puts $socket "PONG $origin"
	}
}

namespace eval ui {
	proc chat {id} {
		set top .irc-[join $id -]
		toplevel $top
		foreach x {row column} { grid ${x}configure $top 0 -weight 1 }
		grid [text $top.in] -sticky news -padx 2 -pady 1
		grid [entry $top.out] -sticky ew -padx 2 -pady 1
		bind $top.out <Return> [nc send $id %W]
	}
	method IRC-Channel incoming {from text} {
		upvar 1 self self
		foreach {nick from} [split $from !] break
		set top .irc-[join $self -]
		if {![winfo exists $top]} { chat $self }
		$top.in insert end "[clock format [clock seconds] -format %X] $nick: $text\n"
	}

	proc send {self w} {
		set txt [$w get]
		$self send $txt
		$w delete 0 end
		incoming Me $txt
	}
}

tags::new {IRC kangaroo} {IRC-Server 0.4} kangaroo.sam-solutions.net
after idle [nc tags::new {{IRC kangaroo} #sam-ru} {IRC-Channel 0.4}]
after idle [nc {IRC kangaroo} connect]

