#  Alicq ICQ client: sound notification module 
#  Copyright (C) Ihar Viarheichyk 2001

#  This module is a part of Alicq ICQ client.

package require resource

namespace eval meta {
	set author "Ihar Viarheichyk <iverg@mail.ru>"
	set description "Basic implementation of sound notifications."
	array set player {description "External player" save change}
}

namespace eval player {
	variable players {auplay esdplay artsplay bplay play aplay}
	array set mapping {text Message URL Message authorization Auth }

	resource::class sounds {} *.wav
	resource sounds {Message Auth}

	proc map {type} {
		variable mapping
		variable files
		if {[info exists mapping($type)]} { set type $mapping($type) }
		resource::origin sounds $type
	}

	# Find external player if possible
	proc external_player {var} {
		variable player
		variable players
		if {[set x [External $players]]!=""} {
			set $var $x
			my log info "Found external player $player"
			return 1
		} else { return 0 }
	}
}

proc player::init {} {
	#  Check if snack extension is available
	upvar 0 [namespace parent]::player player
	if {[info exists player] && $player==""} { unset player }
	if {![info exists player] && [info exists ::env(PLAYER)]} {
		set player $::env(PLAYER)
	}
	if {[info exists player] || [catch {package require snack}]} {
		if {![info exists player]&&$::tcl_platform(platform)=="unix"} {
			external_player player 
		}
		if {[info exists player]} {
			my log info "Using external player $player"
		} else { my log warning "No sound player found"}
		proc play {file} { 
			upvar 0 [namespace parent]::player cmd
			if {[info exists cmd]} { exec $cmd $file & }
		}
	} else {
		my log info "Using snack player"
		snack::sound [namespace current]::Sound
		proc play {file} {
			Sound configure -file $file
			Sound play -blocking no
		}
	}
}

# Intercept incoming events for contacts
include Contact [namespace current] 0.4
proc incoming {type args} { catch { player::play [player::map $type] } }
after idle [nc player::init]

