#file This module is a part of Alicq ICQ client
# It is used to highlight URLs in incoming messages
# and invoke proper handler if user clicks on URL.
#
# Author: Ihar Viarheichyk

option add *urlColor blue widgetDefault

# This program names are used to guess default browser in Unix
array set browsers {
	graphical {mozilla firefox konqueror opera netscape galeon}
	text {links lynx}
	terminals {xterm aterm rxvt eterm konsole}
}
namespace eval meta {
	set author "Ihar Viarheichyk <iverg@mail.ru>"
	set description "Highlight and invoke proper browsers for URLs"
	set name "URL handlers"
	set requires dialogs
	set capabilities {UI unloadable}
	set revision [string map {$ "" " " "" Revision: ""} {$Revision: 1.27 $}]
	set iconfile /usr/share/pixmaps/netscape.png
}

# Two kinds of URLs are recognized at the moment. URLs of first type are opened
# in browser (http, https, ftp, gopher), second one is opened in mailer
# (mailto, news).
# Default openers are stored in BROWSER and MAILER environment variables
switch $::tcl_platform(platform) {
	windows { set ::env(BROWSER) start }
	unix {
		if {![info exists ::env(BROWSER)]} {
			# Use default MacOS-X browser	
			if {$::tcl_platform(os)=="Darwin"} { 
				set ::env(BROWSER) open 
			} else {
				# In Unix ther is no default browser, 
				# trying to guess

				# Check if DE is in use
				if {[info exists ::env(DESKTOP_SESSION)]} {
					switch -exact $::env(DESKTOP_SESSION) {
						Gnome {	set ::env(BROWSER)\
								gnome-open
						}
						kde { set ::env(BROWSER)\
							"kfmclient exec"
						}
					}
				}

				# If browser still not found, try to guess
				if {![info exists ::env(BROWSER)]} {
					# Check popular graphical browsers first
					set x [External $browsers(graphical)]
					if {$x!=""} {
						set ::env(BROWSER) $x
					} else {
						# If graphical browser not found try 
						# text ones.
						# First detect terminal program
						if {[set term [External $browsers(terminals)]]!="" &&
						    [set x [External $browsers(text)]]!=""} {
							set ::env(BROWSER) [list $term -e $x]
						}
					}
				}
			}
		}
	}
}
unset browsers

# If no mailer provided, consider broswer can open such URLs as well
if {![info exists ::env(MAILER)]} {
	set ::env(MAILER) $::env(BROWSER)
}

puts "Browser is $::env(BROWSER)"

proc currentUrl txt {
	set tags [$txt tag names current]
	eval $txt get [$txt tag prevrange url "current + 1c"]
}

# Invoke URL handler
proc UrlHandler {txt} {
	set tags [$txt tag names current]
	set url [currentUrl $txt]
	# Determine all possible types for this URL
	set types [list]
	if {[set i [lsearch $tags url:*]]==-1} return

	set opener [lindex [split [lindex $tags $i] :] end]
	if {[info exists ::env($opener)]} {
		set res [catch { eval exec $::env($opener) $url & } reason]
		if {$res} { Event Log error $reason }
	}
}

proc CopyUrl {txt} {
	clipboard clear
	clipboard append [currentUrl $txt]
}

include Chat UrlChat 0.5

# Initialize url-specific tag handling
method UrlChat initialize {args} {
	upvar 1 self self
	set w [$self in]
	$w tag configure url -foreground [option get $w\
			urlColor UrlColor] -underline yes
	$w tag bind url <Enter> {%W configure -cursor hand2}
	$w tag bind url <Leave> {%W configure -cursor {}}
	$w tag bind url <ButtonRelease-1> [nc UrlHandler %W]
	$w tag bind url <3> [nc UrlContext %W %X %Y]
}

menu .urlcontext -type normal -tearoff no
bind .urlcontext <FocusOut> {%W unpost}
.urlcontext add command -label [mc "Open"]
array set font [font actual [option get .urlcontext font Font]]
set font(-weight) bold
.urlcontext entryconfigure 0 -font [array get font]
.urlcontext add command -label "Open with ..."
.urlcontext add command -label [mc "Copy"]

proc UrlContext {w x y} {
	.urlcontext entryconfigure 0 -command [nc UrlHandler $w]
	.urlcontext entryconfigure 2 -command [nc CopyUrl $w ]
	tk_popup .urlcontext $x $y 0
}

# Detect URLs in message and tag them
method UrlChat highlight {from end} {
	upvar 1 self self
	set w [$self in]
	foreach {opener exp}\
		{ BROWSER {(http|https|ftp|gopher|smb):/[^ ]+}
		  BROWSER {www\.[a-zA-Z][a-zA-Z0-9-]*\.[^ ]+}
		  MAILER  {(mailto|news):[^ ]+}} {
		set i $from
		while {[set i [$w search -count l -regexp $exp $i $end]]!={}} {
			$w tag add url $i "$i + $l char"
			$w tag add url:$opener $i "$i + $l char"
			set i [$w index "$i + $l char"]
		}
	}		
	$w tag raise url
}

