# This module is used to create and update WidowMaker dock icon
# It is a part of Alicq instant messenger client
#
# Author: Ihar Viarheichyk

package require resource
option add *iconFont -*-fixed-medium-r-normal--10-* widgetDefault

array set pending [list]

proc ShowHide {} {
	wm state . [expr ([string equal [wm state .] normal])?"withdrawn":"normal"]
}

proc GetMessages {} {
	foreach x [tags::select Pending] { $x popup; break}
	ui::tooltip::close
}

proc MakeTip {c} {
	set lst [map x [tags::select Pending] { $x alias }]
	if {[llength $lst]} {
		return "[mc {New messages from:}]\n[join $lst \n]"
	} else { return "" }
}

proc CreateIcon {} {
	if {[wm iconwindow .]==""} {
		toplevel .icon -class AlicqIcon
		wm iconwindow . .icon
	}
	canvas .icon.c -background black -width 52 -height 52 -relief sunken
	.icon.c create image 26 26 -anchor s -tag icon
	.icon.c create text 26 52 -anchor s -fill white -tag text
	set font [option get . iconFont Font]
	if {$font!=""} { .icon.c itemconfigure text -font $font }

	pack .icon.c
	bind .icon.c <1> {.main.apple unpost}
	.icon.c bind icon <1> [nc GetMessages]
	.icon.c bind icon <3> {.status.m post %X %Y}
	.icon.c bind text <1> [nc ShowHide]
	.icon.c bind text <3> {.main.apple post %X %Y}
	.icon.c bind icon <Enter> [list ui::tooltip::show [nc MakeTip %W]]
	.icon.c bind icon <Leave> ui::tooltip::close
	my-monitor::new-status [my status]
}

namespace eval pending {
	include Pending [namespace current] 0.8
	proc pending {args} {
		if {![info exists flash::scheduled]} { flash::run 0 }
	}

	proc no-pending {} {
		if {[info exists flash::scheduled] && 
		    [llength [tags::select Pending]]==1} {
		    	after cancel $flash::scheduled
			flash::run 1
		    	after cancel $flash::scheduled
			unset flash::scheduled
		}
	}

	namespace eval flash {
		proc run {stage} {
			variable scheduled
			.icon.c itemconfigure icon -image\
				img:[expr {$stage?[my status]:"message"}]
			set scheduled [after 300 [nc run [expr $stage^1]]]
		}
	}
}

namespace eval my-monitor {
	my include [namespace current]
	proc new-status {status} {
		.icon.c itemconfigure text -text [mc $status]
		.icon.c itemconfigure icon -image img:$status
	}
}

# Hide window instead of deletion
wm protocol . WM_DELETE_WINDOW {wm withdraw .}
resource icons {away dnd ffc group invisible message na occ offline online url}

CreateIcon

